"""build-magic

    * Macro - A command to be executed by a CommandRunner.
    * MacroFactory - Generates Macro objects based on the provided commands.
    * CommandRunner - An abstract class for defining methods for executing commands.
"""

__version__ = '0.1.1rc0'

__all__ = [
    'cli',
    'core',
    'macro',
    'output',
    'runner',
    '__version__',
]
